/*
	demo.c

	Version 1.02
	Copyright (c) 1996, 1997, by Global Constructs. All Rights Reserved.

	This source file includes a program demonstrating the use and utility
	of the memlib.lib static library.  The routines in this library will
	use a special memory based file in place of a disk (or other) based
	file in the temporary directory of the operating system.

	The program shows examples of the following five files:

	1.	Normal random access memory file.
	2.	Stacked FIFO memory file.
	3.	Stacked FILO memory file.
	4.	Stacked FIFO memory file with a maximum file size.
	5.	Normal random access memory file synchronized to a disk based file.

	last modified 10/17/97
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef __BORLANDC__
#	include <unistd.h>
#endif
#include <fcntl.h>
#include <sys/stat.h>

#include "memfile.h"

void printError(void)
{
	printf("Error: %s\n", mstrerror(errmem));
	if (errmem)
		exit(1);
}

/*
	We clear the buffer variable to zero before a string read since the string
	is only partially read and that part is not null terminated.
*/
#define clearBuffer	memset(buffer, 0, sizeof(buffer))

/* setup NULL value */
#ifdef NULL
#undef NULL
#endif
#define NULL	((void *) 0)

void main(void)
{
	char buffer[31], *string = "February29November30December31";
	int bufint;
	struct stat statbuf;
	fileInfo *fptr;
	int16 fh1a, fh1b, fh2, fh3, fh4, fh5;
	int32 ret;
	int dfh;

	printf("\nDemonstration program of memory file system.\n\n");
	printf("Copyright (c) 1997, by Global Constructs.  All Rights Reserved.\n");
	printf("\tVersion 1.02\n\n\n");

	/* Begin Demo 1 */
	printf("Demo 1:\n");
	fh1a = mopen("Random", I_RDWR | I_CREAT, M_INORM | M_IREAD | M_IWRITE);
	if (fh1a < 0)
		printError();
	/* Read and write file 1, outputting the results */

	printf("\tWrite the string '%s' to memory file 1a.\n", string);
	ret = mwrite(fh1a, string, strlen(string)+1);
	if (ret < 0)
		printError();

	printf("\tDuplicate file handle 1a to 1b.\n");
	fh1b = mdup(fh1a);
	if (ret < 0)
		printError();

	printf("\tSeek the file handles for reading.\n");
	ret = mseek(fh1a, 20, SEEK_SET);
	if (ret < 0)
		printError();
	ret = mseek(fh1b, -23, SEEK_END);
	if (ret < 0)
		printError();

	clearBuffer;
	ret = mread(fh1b, buffer, 2);
	if (ret < 0)
		printError();
	bufint = atoi(buffer);
	clearBuffer;
	ret = mread(fh1a, buffer, 8);
	if (ret < 0)
		printError();
	printf("\tRead day %d and month %s.\n", bufint, buffer);

	if (meof(fh1a))
		printf("\tFile handle 1a is at the end of the file.\n");
	ret = mstat("Random", &statbuf);
	if (ret < 0)
		printError();

	printf("\tThe file size before truncation = %ld.\n", (int32) statbuf.st_size);
	ret = mtruncate(fh1b, 13);
	if (ret < 0)
		printError();
	fptr = minfo(fh1a);
	if (fptr) {
		printf("\tThe new file size = %ld.\n", fptr->filesize);
		free(fptr);	/* free the allocated memory when finished with it */
	} else
		printError();
	/* End Demo 1 */
	mclose(fh1a);
	mclose(fh1b);

	/* Begin Demo 2 */
	printf("Demo 2:\n");
	/* open a FIFO file. */
	fh2 = mopen(NULL, I_RDWR | I_CREAT, M_IFIFO | M_IREAD | M_IWRITE);
	if (fh2 < 0)
		printError();
	/* open a FILO file. */
	fh3 = mopen("FILO", I_RDWR | I_CREAT, M_IFILO | M_IREAD | M_IWRITE);
	if (fh3 < 0)
		printError();

	printf("\tThe string before manipulations is '%s'.\n", string);
	/* Push data into file 2 */
	/* read string into FIFO file */
	ret = mpush(fh2, &string[0], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh2, &string[8], 2);
	if (ret < 0)
		printError();
	ret = mpush(fh2, &string[10], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh2, &string[18], 2);
	if (ret < 0)
		printError();
	ret = mpush(fh2, &string[20], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh2, &string[28], 2);
	if (ret < 0)
		printError();

	/* Push data into file 3 */
	/* read string into FILO file */
	ret = mpush(fh3, &string[0], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh3, &string[8], 2);
	if (ret < 0)
		printError();
	ret = mpush(fh3, &string[10], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh3, &string[18], 2);
	if (ret < 0)
		printError();
	ret = mpush(fh3, &string[20], 8);
	if (ret < 0)
		printError();
	ret = mpush(fh3, &string[28], 2);
	if (ret < 0)
		printError();

	/* Pull data from FIFO */
	clearBuffer;
	ret = mpeek(fh2, buffer, 8);
	if (ret < 0)
		printError();
	if (strcmp(buffer, "February"))
		printf("Error: FIFO stack order corrupted.\n");
	ret = mpull(fh2, &buffer[0], 8);
	if (ret < 0)
		printError();
	ret = mpull(fh2, &buffer[8], 2);
	if (ret < 0)
		printError();
	ret = mpull(fh2, &buffer[10], 8);
	if (ret < 0)
		printError();
	ret = mpull(fh2, &buffer[18], 2);
	if (ret < 0)
		printError();
	ret = mpull(fh2, &buffer[20], 8);
	if (ret < 0)
		printError();
	ret = mpull(fh2, &buffer[28], 2);
	if (ret < 0)
		printError();
	printf("\tThe string pulled from the FIFO is '%s'.\n", buffer);

	/* Pull data from FILO */
	clearBuffer;
	ret = mpeek(fh3, buffer, 2);
	if (ret < 0)
		printError();
	if (strcmp(buffer, "31"))
		printf("Error: FILO stack order corrupted.\n");
	ret = mpull(fh3, &buffer[0], 2);
	if (ret < 0)
		printError();
	ret = mpull(fh3, &buffer[2], 8);
	if (ret < 0)
		printError();
	ret = mpull(fh3, &buffer[10], 2);
	if (ret < 0)
		printError();
	ret = mpull(fh3, &buffer[12], 8);
	if (ret < 0)
		printError();
	ret = mpull(fh3, &buffer[20], 2);
	if (ret < 0)
		printError();
	ret = mpull(fh3, &buffer[22], 8);
	if (ret < 0)
		printError();
	printf("\tThe string pulled from the FILO is '%s'.\n", buffer);

	/* End Demo 2 */
	mclose(fh3);
	mclose(fh2);

	/* Begin Demo 3 */
	/* Open file 4 */
	printf("Demo 3:\n");
	fh4 = mopen(NULL, I_RDWR | I_CREAT, M_IFIFO | M_IREAD | M_IWRITE);
	if (fh4 < 0)
		printError();

	/* Set the maximum file size to 10 bytes */
	ret = msetsize(fh4, 10);
	if (ret < 0)
		printError();
	fptr = minfo(fh4);
	if (fptr) {
		printf("\tThe maximum file size = %ld bytes.\n", fptr->maxfilesize);
		free(fptr);	/* free the allocated memory when finished with it */
	} else
		printError();

	/* Push data to the file */
	ret = mpush(fh4, string, strlen(string) + 1);
	if (ret < 0)
		printError();
	printf("\tThe string pushed to the file is '%s'.\n", string);

	/* Read the data in the file */
	clearBuffer;
	ret = mpull(fh4, buffer, strlen(string) + 1);
	if (ret < 0)
		printError();
	printf("\tThe file contains                '%s'.\n", buffer);
	/* End Demo 3 */
	mclose(fh4);

	/* Begin Demo 4 */
	printf("Demo 4:\n");
	printf("\tCreate a disk based file named Text.ext.\n");
	dfh = open("Test.ext", O_CREAT | O_RDWR, S_IWRITE | S_IREAD | S_IWUSR | S_IRUSR);
	if (dfh < 0)
		printError();
	/*
		The following line may report 'Comparison result always True' under Netware.
		10/17/97 CAS.
	*/
	if (write(dfh, string, strlen(string) + 1) < 0)
		printError();
	printf("\tThe disk file initially contains '%s'.\n", string);
	close(dfh);

	printf("\tCreate a sync'd memory file.\n");
	fh5 = mopen("Test.ext", I_CREAT|I_RDWR|I_LOAD, M_ISYNC|M_INORM|M_IREAD|M_IWRITE);
	if (fh5 < 0)
		printError();
	ret = mseek(fh5, 10, SEEK_SET);
	if (ret < 0)
		printError();

	printf("\tWrite 'rebmevoN' to the memory file.\n");
	ret = mwrite(fh5, "rebmevoN", 8);
	if (ret < 0)
		printError();
	mclose(fh5);

	dfh = open("Test.ext", O_RDONLY, S_IREAD);
	if (dfh < 0)
		printError();
	/*
		The following line may report 'Comparison result always True' under Netware.
		10/17/97 CAS.
	*/
	if (read(dfh, buffer, strlen(string) + 1) < 0)
		printError();
	close(dfh);
	printf("\tThe disk file now contains       '%s'.\n", buffer);

	/* End Demo 4 */
	printf("\tDelete Text.ext.\n");
	remove("Test.ext");
}


