/*
    memlib.c

    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 10/17/97

	The functions are external but we do not want just anyone to access them
	directly, so their names are abbreviated with a 'GC' in front.  Source
	contains the following:

		getfilefrompath			gcgffp
		syncfile				gcsyfi
		getfilehandle			gcgefh
		addfilehandle			gcadfh
		addmemorysegment		gcadms
		getsegmentpositionread	gcgesp
		getsegmentpositionwrite	gcgesp
*/
#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include STD_H
#include IO_H
#include STRINGS_H
#include "memlib.h"

static int8u codeformat[18] = GC_CODEFORMAT;

fileList *gcgffp(const char *name)
{
    fileList *ptr;

    if (name) {
        for (ptr=firstlist;ptr;ptr=ptr->f_next)
			/* PTW: Only if there is a filename. 1-14-1997 */
			if(ptr->filename)
            	if (strcmp(ptr->filename, name) == 0)
                	return ptr;
    } else
		ptr = (fileList *) codeformat;
    return NULL;
}

int gcsyfi(fileList *fptr)
{
	memList *mptr;
	int i;

	if (fptr->syncrate) {
		fptr->synczero--;
		if (!fptr->synczero) {
			fptr->synczero = fptr->syncrate;
			mptr = fptr->memsegments;
			if (fptr->filesize) {
				if (lseek((int)fptr->synchandle, 0, SEEK_SET) == -1) {
					errmem = syncSeekFailure;
					return (-1);
				}
				/* partial first block read for FIFO file */
				if (fptr->filemode & M_IFIFO)
					i = write((int)fptr->synchandle, mptr->m_ptr + fptr->startpos, mptr->fblock_size - fptr->startpos);
				else
					i = write((int)fptr->synchandle, mptr->m_ptr, mptr->fblock_size);
				if (i <= 0) {
					errmem = syncWriteFailure;
					return (-1);
				}
				for(mptr=mptr->m_next;mptr;mptr=mptr->m_next) {
					if (mptr->fblock_size) {
						i = write((int)fptr->synchandle, mptr->m_ptr, mptr->fblock_size);
						if (i <= 0) {
							errmem = syncWriteFailure;
							return (-1);
						}
					}
				}
			}
			/* change file size to match memory file */
			if (ftruncate(fptr->synchandle, fptr->filesize) < 0)
				return (-1);
		}
	}
	return 0;
}

fileHandle *gcgefh(int16 handle)
{
	fileHandle *ptr;

	for(ptr=firsthandle;ptr;ptr=ptr->h_next)
		if (ptr->filehandle == handle)
			return ptr;
	return NULL;
}

int16 gcadfh(fileList *fptr, int16 access)
{
	static int16 nextfilehandle;
	fileHandle *hptr;

	/* allocate a file handle */
	hptr = calloc(1, sizeof(fileHandle));
	if (hptr == NULL) {
		errmem = outOfMemory;
		return (-1);
	}
	/* add file handle to beginning of file handle list */
	hptr->h_next = firsthandle;
	hptr->fileaccess = access;
	hptr->file = fptr;

	/* we set this to reduce the test in getsegmentpointer() for FIFO/FILO */
	if (hptr->file->filemode & (M_IFIFO | M_IFILO))
		hptr->last.filepos = -1;

	firsthandle = hptr;
	if (nextfilehandle < 0) {
		/*
			By incrementing 'nextfilehandle', int has overflowed.  To get a file
			handle now we must search file handle list for the next free one.
		*/
		fileHandle *fhptr;
		int16 i = 0;

		do {
			/* start here AFTER the file handle we're adding */
			fhptr = firsthandle->h_next;
			if (fhptr == NULL) {
				/* The file handle list is empty so reset nextfilehandle */
				nextfilehandle = 0;
				return (hptr->filehandle = nextfilehandle++);
			} else
				for(;fhptr;fhptr=fhptr->h_next) {
					if (fhptr->filehandle == i)
						continue;
					else
						return (hptr->filehandle = i);
				}
		} while (++i > 0);
		firsthandle = hptr->h_next;
		free(hptr);
		errmem = tooManyOpenHandles;
		return (-1);
	} else
		return (hptr->filehandle = nextfilehandle++);
}

memList *gcadms(fileList *fptr)
{
	memList *ptr;
	int32 calcsize;

	/*
		Calculate the size that memList should take up so that the pointer at
		the end will point to m_size AND be an aligned pointer. 10297, CAS.
	*/
	/*
		The following line of code will often produce an 'Unreachable Code'
		warning on many platforms.  This is NORMAL, and this test has been
		implemented this way for compatibility with other platforms and
		because the use of sizeof() on an #if line is not ANSI compliant.
		10/17/97 CAS.
	*/
	if (sizeof(memList) % sizeof(memList *)) {
		calcsize = sizeof(memList *) * (sizeof(memList) / sizeof(memList *) + 1);
	} else
		calcsize = sizeof(memList);
	ptr = malloc(calcsize + gcmemheap);
	if (ptr == NULL) {
		errmem = outOfMemory;
		return NULL;
	}
	ptr->m_next = NULL;
	ptr->m_ptr = (char *) ptr + calcsize;
	ptr->m_size = gcmemheap;
	ptr->fblock_size = 0;
	ptr->free_size = ptr->m_size;
	/* attach memory structure to end of file's memory list */
	if (fptr->memsegments) {
		/* set up current last position to point to new block */
		fptr->memsegments->m_prev->m_next = ptr;
		/* set up previous position from new block to old last position */
		ptr->m_prev = fptr->memsegments->m_prev;
		/* set up previous (end) position at first block to new block */
		fptr->memsegments->m_prev = ptr;
	} else
		ptr->m_prev = fptr->memsegments = ptr;

	fptr->clustersize += gcmemheap;
	fptr->numsegments++;
	if (codeformat[7] == 0xC3) {
		return ptr;
	} else {
		return NULL;
	}
}

memList *gcgesp(const fileHandle *hptr, int32 *pos, int readflag)
{
	register memList *mptr;

	if (hptr->filepos == hptr->last.filepos && hptr->last.fpptr) {
		*pos = hptr->last.blockpos;
		return hptr->last.fpptr;
	}

	if (hptr->filepos <= (hptr->file->filesize >> 1)) {
		/* FIFO files will always go here */
		/* start at beginning of file */
		*pos = hptr->filepos + (readflag ? hptr->file->startpos : 0);
		/* an alternate to the for loop (slightly faster) */
		mptr = hptr->file->memsegments;
		while(1) {
			if (*pos >= mptr->m_size)
				*pos -= mptr->m_size;
			else
				return mptr;
			if (mptr->m_next) {
				mptr = mptr->m_next;
				continue;
			}
			break;
		}
	} else {
		/* FILO files will go here when the file size is > zero */
		/* start at end of file */
		*pos = hptr->file->filesize - hptr->filepos;
		if (*pos >= 0) {
			mptr = hptr->file->memsegments->m_prev;
			do {
				if (*pos <= mptr->fblock_size) {
					*pos = mptr->fblock_size - *pos;
					if (*pos == mptr->m_size)
						return NULL;
					else
						return mptr;
				} else
					*pos -= mptr->fblock_size;
				mptr = mptr->m_prev;
			} while (mptr->m_next != NULL);
		}
	}
	return NULL;
}

/* Load ftruncate() function as necessary.  10/17/97 CAS */
#if defined(__TURBOC__) || defined(__WATCOMC__)
#	if defined(__WINDOWS__) || defined(__NT__) || defined(_Windows)
#		include "ftrunc.c"
#	elif defined(__DOS__) || defined(__MSDOS__)
#		include "ftrunc.c"
#	elif defined(__NETWARE_386__)
#		include "ftrunc.c"
#	elif defined(__OS2__)
#		include "ftrunc.c"
#	endif
#elif defined(__BSD__)
	/* Perform no action here */
#elif defined(__MWERKS__)
#	include "ftrunc.c"
#endif

