/*
    mopen.c

	Version 2.05
    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 10/17/97
*/
#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include STD_H
#include FCNTL_H
#include IO_H
#include STRINGS_H
#include "memlib.h"

/*
	This copyright notice must remain in all copies of the source code, whether
	registered, freeware, or otherwise.  Every effort by the user must be taken
	to insure that no compilation, optimization, or modification removes this
	copyright notice from the final executable version of the source code.

	Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.
*/
static char *copr = "Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.";
static int32u version = 0x00020005; /* version 2.05 */

/* size of memory space to allocate for create and write calls. */
int32 gcmemheap = DEFAULT_MHEAP;

fileList *gc1stlist = NULL;  /* pointer to file info list */
int16 errmem = 0;

#define DEFAULT_SYNC	100	/* # of write operations before write to disk */

int16 mopen(const char *path, int16 access, int16u mode)
{
	char *buffer;
	fileList *ptr;
	static int32u qbuf[4] = {0xC7939682,0x8193C396,0x95A2A399,0xA483A3A2};
	int32 len;
	int16 mfh;
	int fh;

	/* check for read and write accesses only */
	switch (access & (I_RDONLY | I_WRONLY | I_RDWR)) {
		case I_RDONLY: case I_WRONLY: case I_RDWR:
			break;
		case 0:
			/* prevents unreferenced warning */
			len = strlen(copr) + version;
			/* no file access specified so default to read only */
			access |= I_RDONLY;
			break;
		default:
			errmem = accessModeConflict;
			return (-1);
		}
	/* look for preexisting memory file */
	if ((ptr = getfilefrompath(path)) != NULL) {
		if (access & I_EXCL) {
			errmem = (access & I_CREAT ? exclCreateFailed : exclAccessDenied);
			return (-1);
		}
		if (access & I_LOAD) {
			errmem = loadFailFileExists;
			return (-1);
		}
		/*
			Permit reading on readable files and writing on writable files
			only.
		*/
		if (access & I_RDONLY && NOT(ptr->filemode & M_IREAD)) {
			errmem = fileNotReadable;
			return (-1);
		} else if (access & (I_WRONLY | I_RDWR) && NOT(ptr->filemode & M_IWRITE)) {
			errmem = fileNotWritable;
			return (-1);
		}
		/* check that already opened file is not a stacked file */
		if (ptr->filemode & (M_IFIFO | M_IFILO)) {
			fileHandle *tmptr;
			/*
				If file is stacked then check how it's currently opened.
				Permit one I_RDONLY and one I_WRONLY file handles.  There can
				be only one file handle if I_RDWR is used.
			*/
			for (tmptr=firsthandle;tmptr;tmptr=tmptr->h_next) {
				if (tmptr->file != ptr)
					continue;
				if ((tmptr->fileaccess & I_RDWR) ||
					((tmptr->fileaccess & I_RDONLY) && (access & I_RDONLY)) ||
					((tmptr->fileaccess & I_WRONLY) && (access & I_WRONLY)))
				{
					errmem = stackFileDup;
					return (-1);
				}
			}
		}
		if ((mfh = addfilehandle(ptr, (access | I_BINARY))) < 0)
			return (-1);
		if (access & I_TRUNC) {
			if (ptr->filemode & M_IWRITE) {
				if (mtruncate(mfh, 0) < 0)
					return (-1);
			} else {
				errmem = noTruncOnRead;
				return (-1);
			}
		}
        return mfh;
	}
	/* File is not already in memory so check disk and load into memory. */
	/*
		PTW: fh eventually gets assigned without question.
		Make sure it gets initialized to a known value. 1-16-1997
	*/
	fh = 0;
	if (access & I_LOAD) {
		fh = open(path, O_RDWR);
		if (fh < 0){
			/* errmem = loadFailModeConflict; */
			return (-1);
		}
		access |= I_CREAT;
	} else if (NOT(access & I_CREAT)) { /* check for I_CREAT being set */
		errmem = fileDoesNotExist;
		return (-1);
	}
	/* check for normal, fifo, or filo modes only */
	switch (mode & (M_IFIFO | M_IFILO | M_INORM)) {
		case M_IFIFO: case M_IFILO: case M_INORM:
			break;
		case 0:
			/* no file type is set so default to normal */
			mode |= M_INORM;
			break;
		default:
			if (access & I_LOAD) close (fh);
			errmem = fileModeConflict;
			return (-1);
	}
	if (mode & M_ISYNC && NOT(access & I_LOAD)) {
		errmem = syncDiskFileOnly;
		return (-1);
	}
	/*
		Permit reading on created readable files and writing on created
		writable files only.
	*/
	if (access & I_RDONLY && NOT(mode & M_IREAD)) {
		if (access & I_LOAD) close (fh);
		errmem = fileNotReadable;
		return (-1);
	} else if (access & (I_WRONLY | I_RDWR) && NOT(mode & M_IWRITE)) {
		if (access & I_LOAD) close (fh);
		errmem = fileNotWritable;
		return (-1);
	}
	/* allocate space for structure filelist and link to list */
	ptr = calloc(1, sizeof(fileList));
	if (ptr == NULL) {
		if (access & I_LOAD) close (fh);
		errmem = outOfMemory;
		return (-1);
	}
	/* add to the beginning of the list */
	ptr->f_next = firstlist;
	firstlist = ptr;
	/*
		Allocate space for filename.  The NULL path pointer is handled this way
		so that if we add features which change the filename then only one file
		will be affected.
	*/
	/* PTW: If path is NULL, don't allocate memory for filename. 1-14-1997 */
	if(path != NULL){
		ptr->filename = malloc(strlen(path)+1);
		if (ptr->filename == NULL) {
			firstlist = ptr->f_next;
			free(ptr);	/* free filelist memory block */
			if (access & I_LOAD) close (fh);
			errmem = outOfMemory;
			return (-1);
		}
		strcpy(ptr->filename, path);
	}else{
		/* PTW: If path is NULL, make filename NULL too. 1-14-1997 */
		ptr->filename = NULL;
	}
	/*
		PTW: Initialise the file's reference buffer. 1-24-1997
	*/
	memcpy(ptr->rbuf,qbuf,sizeof(qbuf));
	ptr->filemode = mode;
	if (ptr->filemode & M_ISYNC) {
		ptr->syncrate = ptr->synczero = DEFAULT_SYNC;
		ptr->synchandle = fh;
	}
	/* allocate memory list for file */
	if (addmemorysegment(ptr) == NULL) {
		if (access & I_LOAD) close (fh);
		/* PTW: Only free it if it's there. 1-14-1997 */
		if(ptr->filename != NULL) free(ptr->filename);
		firstlist = ptr->f_next;
		free(ptr);
		return (-1);
	}
	if ((mfh = addfilehandle(ptr, (access | I_BINARY))) < 0) {
		if (access & I_LOAD) close (fh);
		/*
			removed free(m_ptr) since this is part of the same memory block as
			memList now. 10297, CAS
		*/
		free(ptr->memsegments);
		/* PTW: Only free it if it's there. 1-14-1997 */
		if(ptr->filename != NULL) free(ptr->filename);
		firstlist = ptr->f_next;
		free(ptr);
		return (-1);
	}
	if (access & I_LOAD) {
		buffer = calloc(1,DEFAULT_MHEAP);
		if (buffer == NULL) {
			if (access & I_LOAD) close (fh);
			errmem = outOfMemory;
			return (-1);
		}
		/* read in entire disk file */
		/*
			PTW: Changed loop to remove the use of eof().
			This call is not always available on non-DOS systems
			and is not efficient to implement. 1-12-1997
		*/
		while(1) { 
			len = read(fh, buffer, sizeof(buffer));
			/* PTW: len should be 0 if EOF is reached. 1-12-1997 */
			if (len == 0) break;
			if (len < 0) {
				errmem = diskReadError;
				break; /* return with memory file intact */
			}
			len = mwrite(mfh, buffer, len);
			if (len < 0)
				break; /* return with memory file intact */
		}
		mseek(mfh, 0, SEEK_SET);	/* set file position to beginning of file */
		if (NOT(ptr->filemode & M_ISYNC))
			close (fh);
		free(buffer);
	}
    /* setup file handle and access */
    return mfh;
}


