/*
    mpeek.c

    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 1/7/97
*/
#include <stdlib.h>
#include "common.h"
#include STRINGS_H
#include "memlib.h"

int32 mpeek( int16 handle, void *buffer, int32u len )
{
	fileHandle *hptr;
	register memList *mptr;
	char *bufptr;
	int32 oldfilepos;
	int bytesread;
	register int32 blockread;	/* amount of bytes to write to this memory segment */
	int32 blockpos;	/* file position within the current memory segment */

	/* get file list pointer */
	hptr = getfilehandle(handle);
	if (hptr == NULL) {
		errmem = badFileHandle;
		return (-1);
	}
	if (buffer == NULL) {
		errmem = invalidBufferPtr;
		return (-1);
	}
	if (hptr->fileaccess & I_WRONLY) {
		errmem = fileNotReadable;
		return (-1);
	}
	/* return if length to be read is zero */
	if (len == 0)
		return 0;
	if (hptr->file->filesize == 0) {
		hptr->eof = TRUE;
		return 0;
	}
	if (len >= hptr->file->filesize) {
		hptr->eof = TRUE;
		len = hptr->file->filesize;
	}
	oldfilepos = hptr->filepos;
	if (hptr->file->filemode & M_IFIFO) {
		/* set file pointer towards beginning of file */
		hptr->filepos = 0;
	} else if (hptr->file->filemode & M_IFILO) {
		/* set file pointer towards end of file */
		hptr->filepos = hptr->file->filesize - len;
		if (hptr->filepos < 0) hptr->filepos = 0;
	}
	/*
		Now we read the memory file into 'buffer'.  It must be started at
		the file position and span memory segments as required.  If the file
		position is past the last memory segment or the block position is past
		the end of the file then exit.
	*/
	mptr = getsegmentpositionread(hptr, &blockpos);
	if (mptr == NULL) {
		hptr->filepos = oldfilepos;
		return 0;
	}
	bytesread = 0;
	/* save this info for next real read/pull */
	if (hptr->file->filemode & M_INORM) {
		hptr->last.filepos = hptr->filepos;
		hptr->last.blockpos = blockpos;
		hptr->last.fpptr = mptr;
	}

	bufptr = buffer;
	for(;;mptr = mptr->m_next) {
		/*
			Block to be read fits in this one memory segment or this is the last
			memory segment for this file.
		*/
		blockread = mptr->m_size - blockpos;
		if (blockread >= len || mptr->m_next == NULL) {
			/* read the length requested or the remaining block, whichever is smaller */
			len = min(mptr->fblock_size - blockpos, len);
			/* check for end of file */
			if (hptr->file->filemode & (M_IFIFO | M_IFILO)) {
				if (len == hptr->file->filesize)
					hptr->eof = TRUE;
			} else {
				if (len == mptr->fblock_size - blockpos)
					hptr->eof = TRUE;
			}
			memcpy(bufptr, mptr->m_ptr + blockpos, len);
			hptr->filepos = oldfilepos;
			return bytesread + len;
		}
		/* read the remaining buffer information */
		memcpy(bufptr, mptr->m_ptr + blockpos, blockread);
		bufptr += blockread;
		len -= blockread;
		bytesread += blockread;
		blockpos = 0;
	}
}


