/*
    mpull.c

    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 1/5/97
*/
#include <stdlib.h>
#include "common.h"
#include STRINGS_H
#include "memlib.h"

int32 mpull( int16 handle, void *buffer, int32u len )
{
	fileHandle *hptr;
	memList *mptr, *tmptr;
	char *bufptr;
	int bytesread = 0;
	int32 blockread,	/* amount of bytes to write in this memory segment */
		blockpos;		/* file position within the current memory segment */

	/* get file list pointer */
	hptr = getfilehandle(handle);
	if (hptr == NULL) {
		errmem = badFileHandle;
		return (-1);
	}
	if (buffer == NULL) {
		errmem = invalidBufferPtr;
		return (-1);
	}
	if (NOT(hptr->file->filemode & (M_IFIFO | M_IFILO))) {
		errmem = pullNotAllowed;
		return (-1);
	}
	if (hptr->fileaccess & I_WRONLY) {
		errmem = fileNotReadable;
		return (-1);
	}
	/* return if length to be read is zero */
	if (len == 0)
		return bytesread;
	if (hptr->file->filesize == 0) {
		hptr->eof = TRUE;
		return bytesread;
	}
	if (len > hptr->file->filesize)
		len = hptr->file->filesize;
	/*
		Now we read the memory file into 'buffer'.  It must be started at the
		file position and span memory segments as required.  For FIFO files
		this is at the beginning file position, for FILO this is at the end of
		the stacked file less the buffer length to be read.
	*/
	if (hptr->file->filemode & M_IFIFO) {
		/* set file pointer towards beginning of file */
		hptr->filepos = 0;
	} else {
		/* set file pointer towards end of file */
		hptr->filepos = hptr->file->filesize - len;
		if (hptr->filepos < 0) hptr->filepos = 0;
	}
	mptr = getsegmentpositionread(hptr, &blockpos);
	if (mptr == NULL)
		return bytesread;
	bufptr = buffer;
	for(;;) {
		/*
			Block to be read fits in this one memory segment or this is the last
			memory segment for this file.
		*/
		blockread = mptr->m_size - blockpos;
		if (blockread >= len || mptr->m_next == NULL) {
			/* read the length requested or the remaining block, whichever is smaller */
			if (hptr->file->filemode & M_IFILO) {
				len = min(mptr->fblock_size - blockpos, len);
			} else {
				len = min(blockread, len);
				hptr->file->startpos += len;
			}
			memcpy(bufptr, mptr->m_ptr + blockpos, len);
			hptr->file->filesize -= bytesread + len;
			if (hptr->file->filesize == 0)
				hptr->eof = TRUE;
			mptr->fblock_size -= len;
			mptr->free_size += len;
			/* let's keep the last memory segment for the file */
			if (mptr->fblock_size == 0) {
				if (hptr->file->filemode & M_IFILO && hptr->file->memsegments != mptr) {
					hptr->file->clustersize -= mptr->m_size;
					hptr->file->numsegments--;
					mptr->m_prev->m_next = mptr->m_next;
					if (mptr->m_next)
						mptr->m_next->m_prev = mptr->m_prev;
					else
						hptr->file->memsegments->m_prev = mptr->m_prev;
					/* free memory segment block 10297, CAS */
					free(mptr);		/* free memory segment structure */
				}
				else if (hptr->file->filemode & M_IFIFO && mptr->m_next) {
					hptr->file->clustersize -= mptr->m_size;
					hptr->file->numsegments--;
					mptr->m_next->m_prev = hptr->file->memsegments->m_prev;
					hptr->file->memsegments = mptr->m_next;
					if (mptr->m_next == NULL)
						hptr->file->memsegments->m_prev = mptr->m_prev;
					hptr->file->startpos = 0;
					/* free memory segment block removed 10297, CAS */
					free(mptr);			/* free memory segment structure */
				}
			}
			/* set new file position */
			if (hptr->file->filemode & M_IFIFO)
				hptr->filepos = 0;	/* at beginning of file */
			else
				hptr->filepos = hptr->file->filesize;	/* at end of file */
			break;
		}
		/* read the remaining buffer information */
		memcpy(bufptr, mptr->m_ptr + blockpos, blockread);
		bufptr += blockread;
		len -= blockread;
		bytesread += blockread;
		mptr->fblock_size -= blockread;
		mptr->free_size += blockread;
		blockpos = 0;
		if (mptr->fblock_size == 0) {
			hptr->file->clustersize -= mptr->m_size;
			hptr->file->numsegments--;
			if (hptr->file->filemode & M_IFIFO) {
				hptr->file->startpos = 0;
				mptr->m_next->m_prev = hptr->file->memsegments->m_prev;
				hptr->file->memsegments = mptr->m_next;
				/* free memory segment block removed 10297, CAS */
				free(mptr);			/* free memory segment structure */
				mptr = hptr->file->memsegments;
				continue;
			} else {
				if (mptr == hptr->file->memsegments) {
					mptr->m_next->m_prev = hptr->file->memsegments->m_prev;
					tmptr = hptr->file->memsegments = mptr->m_next;
				} else {
					tmptr = mptr->m_prev->m_next = mptr->m_next;
					mptr->m_next->m_prev = mptr->m_prev;
				}
				if (mptr->m_next == NULL)
					hptr->file->memsegments->m_prev = mptr->m_prev;
				/* free memory segment block removed 10297, CAS */
				free(mptr);			/* free memory segment structure */
				mptr = tmptr;
				continue;
			}
		}
		mptr = mptr->m_next;
	}
	syncfile(hptr->file);
	return bytesread + len;
}


