/*
    mread.c

    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 1/7/97
*/
#include <stdlib.h>
#include "common.h"
#include STRINGS_H
#include "memlib.h"

int32 mread( int16 handle, void *buffer, int32u len )
{
	fileHandle *hptr;
	register memList *mptr;
	char *bufptr;
	int bytesread;
	register int32 blockread;	/* amount of bytes to write in this memory segment */
	int32 blockpos;		/* file position within the current memory segment */

	/* get file list pointer */
	hptr = getfilehandle(handle);
	if (hptr == NULL) {
		errmem = badFileHandle;
		return (-1);
	}
	if (buffer == NULL) {
		errmem = invalidBufferPtr;
		return (-1);
	}
	if (hptr->file->filemode & (M_IFIFO | M_IFILO)) {
		errmem = stackFileRead;
		return (-1);
	}
	if (hptr->fileaccess & I_WRONLY) {
		errmem = fileNotReadable;
		return (-1);
	}
	/* return if length to be read is zero */
	if (len == 0 || meof(handle))
		return 0;
	/*
		Now we read the memory file into 'buffer'.  It must be started at
		the file position and span memory segments as required.  If the file
		position is past the last memory segment or the block position is past
		the end of the file then exit.
	*/
	mptr = getsegmentpositionread(hptr, &blockpos);
	if (mptr == NULL)
		return 0;
	bufptr = buffer;
	bytesread = 0;
	for(;;mptr=mptr->m_next) {
		/*
			Block to be read fits in this one memory segment or this is the last
			memory segment for this file.
		*/
		blockread = mptr->m_size - blockpos;
		if (blockread >= len || mptr->m_next == NULL) {
			/* read the length requested or the remaining block, whichever is smaller */
			len = min(mptr->fblock_size - blockpos, len);
			/* check for end of file */
			if (len == mptr->fblock_size - blockpos)
				hptr->eof = TRUE;
			memcpy(bufptr, mptr->m_ptr + blockpos, len);
			hptr->filepos += bytesread + len;
			hptr->last.filepos = hptr->filepos;
			hptr->last.blockpos = blockpos + len;
			if (mptr->m_size == hptr->last.blockpos) {
				hptr->last.blockpos = 0;
				hptr->last.fpptr = mptr->m_next;
			} else
				hptr->last.fpptr = mptr;
			return bytesread + len;
		}
		/* read the remaining buffer information */
		memcpy(bufptr, mptr->m_ptr + blockpos, blockread);
		bufptr += blockread;
		len -= blockread;
		bytesread += blockread;
		blockpos = 0;
	}
}


