/*
    mwrite.c

    Copyright (c) 1996, 1997, by Global Constructs.  All Rights Reserved.

    last modified 10/17/97
*/
#include "common.h"
#include <stdio.h>
#include STRINGS_H
#include "memlib.h"

static int zerotofilepos(fileHandle *hptr);

int32 mwrite( int16 handle, void *buffer, int32u len )
{
	fileHandle *hptr;
	register memList *mptr;
	char *bufptr;
	int byteswritten;
	register int32 blockwrite;	/* amount of bytes to write in this memory segment */
	int32 blockpos;	/* file position within the current memory segment */

	/* get file list pointer */
	hptr = getfilehandle(handle);
	if (hptr == NULL) {
		errmem = badFileHandle;
		return (-1);
	}
	if (buffer == NULL) {
		errmem = invalidBufferPtr;
		return (-1);
	}
	if (hptr->fileaccess & I_RDONLY) {
		errmem = fileNotWritable;
		return (-1);
	}
	/* return if length to be written is zero */
	if (len == 0)
		return 0;
	/* clear EOF on write */
	hptr->eof = FALSE;
	/*
		Check for file position past the end of file and comply with POSIX
		resolution.
	*/
	if ((hptr->file->filemode & (M_IFIFO | M_IFILO)) || (hptr->fileaccess & I_APPEND))
		hptr->filepos = hptr->file->filesize;
	else if (zerotofilepos(hptr))
		return 0;
	/* Check and limit the length of the write if maxfilesize is set */
	if (hptr->maxfilesize) {
		if (hptr->maxfilesize > hptr->file->filesize) {
			/*
				We know that (filepos NOT > (filesize AND maxfilesize)) because
				it would have failed in zerotofilepos() and returned.
			*/
			if (hptr->filepos + len > hptr->maxfilesize) {
				len = hptr->maxfilesize - hptr->filepos;
				errmem = maxFileSizeExceeded;
			}
		} else {
			if (hptr->filepos + len > hptr->file->filesize) {
				len = hptr->file->filesize - hptr->filepos;
				errmem = maxFileSizeExceeded;
			}
		}
	}
	if (len == 0)
		return 0;
    /*
        Now we write the contents of 'buffer' to the memory file.  It must be
		started at the file position and add additional memory segments as space
		is needed.
    */
    mptr = getsegmentpositionwrite(hptr, &blockpos);
    if (mptr == NULL) {
        mptr = addmemorysegment(hptr->file);
        if (mptr == NULL)
            return (-1);
        blockpos = 0;
    }
	byteswritten = 0;
    bufptr = buffer;
    for(;;mptr=mptr->m_next) {
        /* block to be written fits in this one memory segment */
        blockwrite = mptr->m_size - blockpos;
        if (len <= blockwrite) {
            memcpy(mptr->m_ptr + blockpos, bufptr, len);
            /*
                If the memory segment has free space then we need to check if
				the block we just copied exceeded the end of file.  Remember
				that free space can only occur in the last memory segment for
				the file.
			*/
			if (mptr->free_size > 0)
				if ((blockpos + len) > mptr->fblock_size) {
					hptr->file->filesize += blockpos + len - mptr->fblock_size;
					mptr->fblock_size = blockpos + len;
					mptr->free_size = mptr->m_size - mptr->fblock_size;
				}
			hptr->filepos += len;
			if (hptr->filepos == hptr->file->filesize)
				hptr->eof = TRUE;
			break;
		}
		/* fill the remaining block */
		memcpy(mptr->m_ptr + blockpos, bufptr, blockwrite);
		bufptr += blockwrite;
		len -= blockwrite;
		byteswritten += blockwrite;
		/* See if we wrote past the end of the file. */
		if (mptr->free_size > 0)
			if ((blockpos + blockwrite) > mptr->fblock_size)
				hptr->file->filesize += mptr->free_size;
		mptr->fblock_size = mptr->m_size;
		mptr->free_size = 0;
		hptr->filepos += blockwrite;
		/* check for next memory segment */
		if (mptr->m_next) {
			blockpos = 0;
			continue;
		}
		/* no remaining memory segments so create one */
		if (addmemorysegment(hptr->file) == NULL) {
			if (hptr->filepos == hptr->file->filesize)
				hptr->eof = TRUE;
			return (-1);
		}
		blockpos = 0;
	}
	if (hptr->file->filemode & M_INORM) {
		hptr->last.filepos = hptr->filepos;
		hptr->last.blockpos = blockpos + len;
		if (mptr->m_size == hptr->last.blockpos) {
			hptr->last.blockpos = 0;
			hptr->last.fpptr = mptr->m_next;
		} else
			hptr->last.fpptr = mptr;
	}
	syncfile(hptr->file);
	return byteswritten + len;
}

/*
	static int zerotofilepos (fileHandle *hptr)

	Clear past end of file up to file position.

	Parameter
	hptr	Pointer to file handle list

	Returns zero if successful or no action performed, otherwise -1 is returned
	and errmem is set.

	Errors
	maxFileSizeExceeded
	outOfMemory

	KBLs	This is compliant with the POSIX definition but it is not handled as
			the Unix file system does.  Space is filled with zeroes where the
			Unix file system does not allocate that space for the file but returns
			zeroes if read from that locale.
*/
static int zerotofilepos(fileHandle *hptr)
{
	int flag = 0;	/* did we exceed maximum file size */
	int32u len;
	memList *mptr;

	if (hptr->filepos <= hptr->file->filesize)
		return 0;
	/* check maximum file size limit before extending */
	if (hptr->maxfilesize) {
		if (hptr->maxfilesize > hptr->file->filesize) {
			if (hptr->filepos >= hptr->maxfilesize) {
				len = hptr->maxfilesize - hptr->file->filesize;
				errmem = maxFileSizeExceeded;
				flag = -1;
			} else
				len = hptr->filepos - hptr->file->filesize;
		} else {
			len = 0;
			errmem = maxFileSizeExceeded;
			flag = -1;
		}
	} else
		len = hptr->filepos - hptr->file->filesize;
	if (len == 0)
		return flag;
	/* This loop will keep adding segments until the file position is reached */
	for (mptr = hptr->file->memsegments;;mptr = mptr->m_next) {
		if (mptr->free_size > 0) {
			if (len <= mptr->free_size) {
				memset(mptr->m_ptr + mptr->fblock_size, 0, len);
				mptr->fblock_size += len;
				mptr->free_size -= len;
				hptr->file->filesize += len;
				if (hptr->file->filemode & M_INORM) {
					hptr->last.filepos = hptr->filepos;
					hptr->last.blockpos = mptr->fblock_size + len;
					if (mptr->m_size == hptr->last.blockpos) {
						hptr->last.blockpos = 0;
						hptr->last.fpptr = mptr->m_next;
					} else
						hptr->last.fpptr = mptr;
				}
				return flag;
			}
			/* fill the remaining block */
			memset(mptr->m_ptr + mptr->fblock_size, 0, mptr->free_size);
			len -= mptr->free_size;
			mptr->fblock_size = mptr->m_size;
			hptr->file->filesize += mptr->free_size;
			mptr->free_size = 0;
		}
		/* check for next memory segment */
		if (mptr->m_next)
			continue;
		/* no remaining memory segments so create one */
		if (addmemorysegment(hptr->file) == NULL)
			return (-1);
	}
}


